/***********************************************************************
// Copyright (C) 2004, 2006 International Business Machines and others.
// All Rights Reserved.
// This code is published under the Common Public License.
//
// $Id: IpReturnCodes.h 759 2006-07-07 03:07:08Z andreasw $
//
// Authors:  Carl Laird, Andreas Waechter     IBM    2004-08-13
************************************************************************/

#ifndef __IPRETURNCODES_H__
#define __IPRETURNCODES_H__

/* include from a common include file */

#include "IpReturnCodes_inc.h"

#endif
