// Copyright (C) 2004, 2006 International Business Machines and others.
// All Rights Reserved.
// This code is published under the Common Public License.
//
// $Id: IpTypes.hpp 759 2006-07-07 03:07:08Z andreasw $
//
// Authors:  Carl Laird, Andreas Waechter     IBM    2004-08-13

#ifndef __IPTYPES_HPP__
#define __IPTYPES_HPP__

#include "SimTKcommon/internal/common.h"

namespace SimTKIpopt
{
  /** Type of all numbers */
  typedef SimTK::Real Number;
  /** Type of all indices of vectors, matrices etc */
  typedef int Index;
  /** Type of default integer */
  typedef int Int;

} // namespace Ipopt

#endif
