/***********************************************************************
// Copyright (C) 2006 International Business Machines and others.
// All Rights Reserved.
// This code is published under the Common Public License.
//
// $Id: IpReturnCodes.hpp 759 2006-07-07 03:07:08Z andreasw $
//
// Authors:  Andreas Waechter               IBM    2006-03-01
************************************************************************/

#ifndef __IPRETURNCODES_HPP__
#define __IPRETURNCODES_HPP__

/* include from a common include file */

namespace SimTKIpopt
{
#include "IpReturnCodes_inc.h"
}

#endif
